"""merged assignment configuration options into assignment settings column

Revision ID: fc5d2febe781
Revises: a0718dae969d
Create Date: 2025-01-27 15:42:24.363658

"""
import datetime
import json
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'fc5d2febe781'
down_revision = 'a0718dae969d'
branch_labels = None
depends_on = None

def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()
    raise TypeError ("Type %s not serializable" % type(obj))


def upgrade():
    # Migrate data from old columns to the settings column
    conn = op.get_bind()
    assignments = conn.execute(sa.text("SELECT id, duedate, type, automatic_grading, max_submissions, allow_files, settings FROM assignment"))
    
    for assignment in assignments:
        assignment = dict(zip(assignments.keys(), assignment))
        existing_settings = json.loads(assignment['settings']) if assignment['settings'] else {}
        new_settings = {
            "deadline": assignment['duedate'] if assignment['duedate'] else None,
            "assignment_type": assignment['type'],
            "autograde_type": assignment['automatic_grading'],
            "max_submissions": assignment['max_submissions'],
            "allowed_files": ["*"] if bool(assignment['allow_files']) else []
        }
        existing_settings.update(new_settings)
        conn.execute(
            sa.text("UPDATE assignment SET settings = :settings WHERE id = :id"),
            {"settings": json.dumps(existing_settings, default=json_serial), "id": assignment['id']}
        )

    # Drop the old columns
    op.drop_column('assignment', 'duedate')
    op.drop_column('assignment', 'type')
    op.drop_column('assignment', 'automatic_grading')
    op.drop_column('assignment', 'max_submissions')
    op.drop_column('assignment', 'allow_files')

def downgrade():
    # Recreate the old columns
    # Drop the old columns
    conn = op.get_bind()
    inspector = sa.inspect(conn)
    columns = [col['name'] for col in inspector.get_columns('assignment')]
    if 'duedate' not in columns:
        op.add_column('assignment', sa.Column('duedate', sa.DateTime(), nullable=True))
    if 'type' not in columns:
        op.add_column('assignment', sa.Column('type', sa.Enum("user", "group", name='assignment_type'), nullable=False, server_default="user"))
    if 'automatic_grading' not in columns:
        op.add_column('assignment', sa.Column('automatic_grading', sa.Enum('unassisted', 'auto', 'full_auto', name='automatic_grading'), server_default='auto', nullable=False))
    if 'max_submissions' not in columns:
        op.add_column('assignment', sa.Column('max_submissions', sa.Integer(), nullable=True, server_default=None, unique=False))
    if 'allow_files' not in columns:
        op.add_column('assignment', sa.Column('allow_files', sa.Boolean(), nullable=False, server_default='f', default=False))

    # Migrate data back from the settings column to the old columns
    assignments = conn.execute(sa.text("SELECT id, settings FROM assignment"))
    
    for assignment in assignments:
        assignment = dict(zip(assignments.keys(), assignment))
        settings = json.loads(assignment['settings']) if assignment['settings'] else {}
        conn.execute(
            sa.text(
                """
                UPDATE assignment
                SET duedate = :duedate, type = :type, automatic_grading = :automatic_grading,
                    max_submissions = :max_submissions, allow_files = :allow_files
                WHERE id = :id
                """
            ),
            {
                "duedate": datetime.datetime.fromisoformat(settings['deadline']) if settings.get('deadline') else None,
                "type": settings.get('assignment_type'),
                "automatic_grading": settings.get('autograde_type'),
                "max_submissions": settings.get('max_submissions'),
                "allow_files": len(settings.get('allowed_files')) > 0,
                "id": assignment['id']
            }
        )


# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import enum
import json
from datetime import date, datetime, timezone
from typing import Any

from grader_service.api.models import assignment
from sqlalchemy import (Column, DateTime, Enum, ForeignKey,
                        Integer, String, Text, Boolean, DECIMAL)
from sqlalchemy.orm import relationship

from grader_service.api.models.assignment_settings import AssignmentSettings
from grader_service.orm.base import Base, DeleteState, Serializable

def get_utc_time():
    return datetime.now(tz=timezone.utc)

def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError ("Type %s not serializable" % type(obj))

class Assignment(Base, Serializable):
    __tablename__ = "assignment"
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(255), nullable=False)
    """Name of the assignment"""
    lectid = Column(Integer, ForeignKey("lecture.id"))
    points = Column(DECIMAL(10, 3), nullable=True)
    status = Column(
        Enum("created", "pushed", "released", "complete"),
        default="created",
    )
    deleted = Column(Enum(DeleteState), nullable=False, unique=False)
    properties = Column(Text, nullable=True, unique=False)
    created_at = Column(DateTime, default=get_utc_time, nullable=False)
    updated_at = Column(DateTime, default=get_utc_time,
                        onupdate=get_utc_time, nullable=False)
    _settings = Column('settings',Text, server_default='', nullable=False)

    lecture = relationship("Lecture", back_populates="assignments")
    submissions = relationship("Submission", back_populates="assignment")

    @property
    def settings(self) -> AssignmentSettings:
        if self._settings is None:
            return AssignmentSettings()
        return AssignmentSettings.from_dict(json.loads(self._settings))
    
    @settings.setter
    def settings(self, settings: AssignmentSettings | dict):
        if isinstance(settings, dict):
            self._settings = json.dumps(settings, default=json_serial)
            return settings
        self._settings = json.dumps(settings.to_dict(), default=json_serial)
        return settings
    
    def update_settings(self, **kwargs: Any):
        # Update specific fields of the AssignmentSettings object
        settings = self.settings  # Get the current AssignmentSettings object
        for key, value in kwargs.items():
            if key not in AssignmentSettings.openapi_types.keys():
                raise RuntimeError(f"provided key '{key}' is not valid for assignment settings")
            if hasattr(settings, key):  # Ensure the attribute exists on AssignmentSettings
                setattr(settings, key, value)
        self.settings = settings  # Save the updated object back

    @property
    def model(self) -> assignment.Assignment:
        assignment_model = assignment.Assignment(
            id=self.id,
            name=self.name,
            status=self.status,
            points=self.points,
            settings=self.settings
        )
        return assignment_model
